/*
  Java gfx-test framework

   Copyright (C) 2010, 2011  Red Hat

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

IcedTea is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
*/

package org.gfxtest.testsuites;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;

import org.gfxtest.framework.*;
import org.gfxtest.framework.annotations.*;



/**
 * This test check the proper rendering of antialiased lines. All sample images
 * and test images are zoomed to show the pixel-wide differences between both
 * renderers. Zoom factor is set to 16.
 * 
 * @author Pavel Tisnovsky
 */
@TestType(TestTypes.RENDER_TEST)
@GraphicsPrimitive(GraphicsPrimitives.LINE)
@RenderStyle(RenderStyles.NORMAL_AA)
@Transformation(Transformations.NONE)
@Zoom(16)
public class AALines extends GfxTest
{
    /**
     * Draw line onto the image using given color and also highlights both line
     * end points.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @param x1
     *            the first point's x coordinate.
     * @param y1
     *            the first point's y coordinate.
     * @param x2
     *            the second point's x coordinate.
     * @param y2
     *            the second point's y coordinate.
     * @param color
     *            line color
     */
    private static void drawLine(BufferedImage image, Graphics2D graphics2d, int x1, int y1, int x2, int y2, Color color)
    {
        // draw the line
        graphics2d.setColor(color);
        graphics2d.drawLine(x1, y1, x2, y2);

        // highlight also both line endpoints
        image.setRGB(x1, y1, 0xff0000);
        image.setRGB(x2, y2, 0xff0000);
    }

    /**
     * Draws various horizontal lines with end points having theirs
     * y-coordinates changed by +- one pixel.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     */
    private static void drawHorizontalJaggedLines(BufferedImage image, Graphics2D graphics)
    {
        // y-difference is +1 pixel
        drawLine(image, graphics,  3,  3, 36,  4, Color.BLACK);
        // y-difference is +1 pixel
        drawLine(image, graphics,  3,  9, 36, 10, Color.YELLOW);
        // y-difference is +1 pixel
        drawLine(image, graphics,  3, 15, 36, 16, Color.MAGENTA);
        // y-difference is -1 pixel
        drawLine(image, graphics,  3, 21, 36, 20, Color.BLUE);
        // y-difference is -1 pixel
        drawLine(image, graphics,  3, 26, 36, 25, Color.GRAY);
    }

    /**
     * Draws various vertical lines with end points having theirs x-coordinates
     * changed by +- one pixel.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     */
    private static void drawVerticalJaggedLines(BufferedImage image, Graphics2D graphics)
    {
        drawLine(image, graphics,  3, 3,  4, 26, Color.BLACK);
        drawLine(image, graphics, 11, 3, 12, 26, Color.YELLOW);
        drawLine(image, graphics, 19, 3, 20, 26, Color.MAGENTA);
        drawLine(image, graphics, 27, 3, 28, 26, Color.BLUE);
        drawLine(image, graphics, 35, 3, 34, 26, Color.GRAY);
    }

    /**
     * Draw color wheel, ie. lines with one common end point and other end point
     * forming circle.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     */
    private static void drawColorWheel(BufferedImage image, Graphics2D graphics)
    {
        // compute center of the image
        int xc = image.getWidth() >> 1;
        int yc = image.getHeight() >> 1;
        // draw color wheel
        for (int i = 0; i < ANGLES; i++)
        {
            // compute color for each line
            Color color = new Color(Color.HSBtoRGB((float)i/ANGLES, 1.0f, 1.0f));
            graphics.setColor(color);
            int majorRadius = (xc > yc ? yc : xc) - 2;
            double angle = 2.0 * i / ANGLES * Math.PI;
            double cos = Math.cos(angle);
            double sin = Math.sin(angle);
            // compute line endpoints
            int x1 = xc + (int)(6.0*cos);
            int y1 = yc + (int)(6.0*sin);
            int x2 = xc + (int)(majorRadius*cos);
            int y2 = yc + (int)(majorRadius*sin);
            drawLine(image, graphics, x1, y1, x2, y2, color);
        }
    }

    /**
     * Draws special types of lines - very short ones with different angles
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     */
    private static void drawSpecialCases(BufferedImage image, Graphics2D graphics)
    {
        int y1 = 2;
        // rows of lines
        for (int j = 0; j < 5; j++)
        {
            int x1 = 2;
            int y2 = y1 + j;
            // columns of lines
            for (int i = 0; i < 6; i++)
            {
                int x2 = x1 + i;
                drawLine(image, graphics, x1, y1, x2, y2, Color.BLACK);
                // next column
                x1 = x2 + 4;
            }
            // next row
            y1 = y2 + 4;
        }
    }

    /**
     * This method just enables antialiasing for the following rendering
     * operation.
     * 
     * @param graphics2d
     *            graphics canvas
     */
    private void setAA1(Graphics2D graphics)
    {
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    }

    /**
     * This method just enables antialiasing and fractional metrics for the
     * following rendering operation.
     * 
     * @param graphics2d
     *            graphics canvas
     */
    private void setAA2(Graphics2D graphics)
    {
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
    }

    /**
     * Draw vertical "jagged" lines with antialiasing disabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testVertStepNoAA(TestImage image, Graphics2D graphics)
    {
        drawVerticalJaggedLines(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Draw horizontal "jagged" lines with antialiasing disabled.
     *
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testHorsStepNoAA(TestImage image, Graphics2D graphics)
    {
        drawHorizontalJaggedLines(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Draw vertical "jagged" lines with antialiasing enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testVertStepAA1(TestImage image, Graphics2D graphics)
    {
        setAA1(graphics);
        drawVerticalJaggedLines(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Draw horizontal "jagged" lines with antialiasing enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testHorsStepAA1(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing
        setAA1(graphics);
        drawHorizontalJaggedLines(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Draw vertical "jagged" lines with antialiasing and fractional metrics
     * enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testVertStepAA2(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing and fractional metrics
        setAA2(graphics);
        drawVerticalJaggedLines(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Draw horizontal "jagged" lines with antialiasing and fractional metrics
     * enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testHorsStepAA2(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing and fractional metrics
        setAA2(graphics);
        drawHorizontalJaggedLines(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Test color wheel (lines with various slope) with antialiasing disabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testColorWheelNoAA(TestImage image, Graphics2D graphics)
    {
        drawColorWheel(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Test color wheel (lines with various slope) with antialiasing enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testColorWheelAA1(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing
        setAA1(graphics);
        drawColorWheel(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Test color wheel (lines with various slope) with antialiasing and
     * fractional metrics enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testColorWheelAA2(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing and fractional metrics
        setAA2(graphics);
        drawColorWheel(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Test various special cases with antialiasing disabled.
     *
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testSpecialCasesNoAA(TestImage image, Graphics2D graphics)
    {
        drawSpecialCases(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Test various special cases with antialiasing enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testSpecialCasesAA1(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing
        setAA1(graphics);
        drawSpecialCases(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Test various special cases with antialiasing and fractional metrics
     * enabled.
     * 
     * @param image
     *            image to which line is to be drawn
     * @param graphics2d
     *            graphics canvas
     * @return test result status - PASSED, FAILED or ERROR
     */
    public TestResult testSpecialCasesAA2(TestImage image, Graphics2D graphics)
    {
        // enable antialiasing and fractional metrics
        setAA2(graphics);
        drawSpecialCases(image.getImage(), graphics);
        return TestResult.PASSED;
    }

    /**
     * Entry point to the test suite.
     *
     * @param args not used in this case
     */
    public static void main(String[] args)
    {
        new AALines().runTestSuite(args);
    }

}
